#include <zephyr/posix/unistd.h>
#include "../includes/batteryCharger.h"
#include <zephyr/device.h>
#include <zephyr/drivers/i2c.h>
#include <zephyr/logging/log.h>

#include "../includes/RTCHelper.h"

#define BATTERY_CHARGER_BUS DT_NODELABEL(i2c2)
#define BATTERY_CHARGER_ADDRESS 107 //I2C address

/* Defualt values for configuration registers, these values are
    set during the init function. Be sure to check the datasheet before
    messing with these.
*/
uint8_t  battery_charger_reg_0x00 = 0x04;    // Minimal system voltage
const uint8_t battery_charger_reg_0x01[2] = {0x02, 0x08}; // Charge Voltage Limit (Currently set to 5.2V)
const uint8_t battery_charger_reg_0x03[2] = {0x00, 0xC8}; // Charge Current Limit
const uint8_t battery_charger_reg_0x06[2] = {0x00, 0xC8}; // Input Voltage Limit Register
uint8_t  battery_charger_reg_0x0A = 0x6F;    // Re-charge Control Register
uint8_t  battery_charger_reg_0x11 = 0x00;    // Charge Control Register 2
uint8_t  battery_charger_reg_0x15 = 0xBA;    // MPPT Control Register
uint8_t  battery_charger_reg_0x2E = 0x40;    // ADC Control Register
    
static const struct device *bChargerDevice = DEVICE_DT_GET(BATTERY_CHARGER_BUS);

//Internal state contaner structure
struct _chargerState _B_CHARGER_INERNAL_STATE = {
    .chargingEnabled = true,
    .mpptEnabled = false,
    .adcEnabled = false
};
struct _chargerState *chargerState = &_B_CHARGER_INERNAL_STATE;

//Function to put the charger IC into a known settings state
void batteryChargerInit(){ //Could be made more efficent by mapping the entire space but isn't super needed
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Writting defualt settings to battery charger!");
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x0A, battery_charger_reg_0x0A); //Contans cell count so must go first
    k_msleep(100); //Let the chip adjsut to the cell change before adjusting rest of registers
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x00, battery_charger_reg_0x00);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x01, battery_charger_reg_0x01[0]);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x02, battery_charger_reg_0x01[1]);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x03, battery_charger_reg_0x03[0]);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x04, battery_charger_reg_0x03[1]);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x06, battery_charger_reg_0x06[0]);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x07, battery_charger_reg_0x06[1]);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x11, battery_charger_reg_0x11);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x15, battery_charger_reg_0x15);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x2E, battery_charger_reg_0x2E);
}

// ADC control function
void batteryCharger_ADC_Trigger(){
    uint8_t buffer;
    i2c_reg_read_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x2E, &buffer);
    i2c_reg_write_byte(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x2E, buffer | 0x80);
    k_msleep(200); //Unoptimized block to let the chip do the conversion.
}

//ADC measurement functions
float batteryCharger_Measure_Battery_Voltage(){
    uint8_t buffer[2];
    i2c_burst_read(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x3B, buffer, 2);
    uint16_t value = ((uint16_t) buffer[0] <<8 | buffer [1]);
    return 0.001 * (float) value;
}
float batteryCharger_Measure_Vsys_Voltage(){
    uint8_t buffer[2];
    i2c_burst_read(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x3D, buffer, 2);
    uint16_t value = ((uint16_t) buffer[0] << 8 | buffer[1]);
    return 0.001 * (float) value;
}
float batteryCharger_Measure_Vac1_Voltage(){
    uint8_t buffer[2];
    i2c_burst_read(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x37, buffer, 2);
    uint16_t value = ((uint16_t) buffer[0] << 8 | buffer[1]);
    return 0.001 * (float) value;
}
float batteryCharger_Measure_Battery_Current(){
    uint8_t buffer[2];
    i2c_burst_read(bChargerDevice, BATTERY_CHARGER_ADDRESS, 0x33, buffer, 2);
    uint16_t value = ((uint16_t) buffer[0] << 0 | buffer[1]);
    return 0.001 * (float) value;
}

//Shell commands
int __shell_Charger_Dump_Memory(const struct shell *sh, size_t argc, char **argv){
    uint8_t chip_memory[72];
    i2c_burst_read(bChargerDevice,BATTERY_CHARGER_ADDRESS, 0x00, chip_memory, 71);
    shell_print(sh, "Printing values of battery charger chip");
    for (int i = 0; i < 72; i++){
        shell_print(sh, "Address: %x - Value: %x ", i, chip_memory[i] & 0xff);
    }
    return 0;
}
int __shell_Charger_Measure_Voltages(const struct shell *sh, size_t argc, char **argv){
    batteryCharger_ADC_Trigger();
    float voltages[3];
    voltages[0] = batteryCharger_Measure_Battery_Voltage();
    voltages[1] = batteryCharger_Measure_Vsys_Voltage();
    voltages[2] = batteryCharger_Measure_Battery_Current();
    shell_print(sh, "Measured as follows: ");
    shell_print(sh, "Vbatt = %1.3f \n Vsys = %1.3f", voltages[0], voltages[1]); 
    shell_print(sh, "Battery Current = %1.4f", voltages[2]);
    return 0;
} 